"""Expose search api from landsat-util."""
from __future__ import absolute_import

import requests

from gaia.core import Time
from gaia import config

_section = 'landsat'

# Set configuration defaults for NASA search API
config.add_section(_section)
config.set(_section, 'url', 'http://landsat.usgs.gov/includes/scripts/get_metadata.php')
config.set(_section, 'datefmt', '%%m/%%d/%%Y')

_sensor_options = {
    'LANDSAT_8': 'Landsat 8 OLI/TIRS',
    'LANDSAT_ETM_SLC_OFF': 'Landsat 7 SLC-off (2003 ->)',
    'LANDSAT_ETM': 'Landsat 7 SLC-on (1999-2003)',
    'LANDSAT_TM': 'Landsat 4-5 TM',
    'LANDSAT_MSS2': 'Landsat 4-5 MSS',
    'LANDSAT_MSS1': 'Landsat 1-3 MSS',
    'LANDSAT_COMBINED': 'Landsat 4-8 Combined'
}

_defaults = {
    'limit': 10,
    'sensor': 'LANDSAT_8',
    'start': '01/01/1990',
    'end': Time(),
    'west': -179.9,
    'east': 179.9,
    'north': 82.6401,
    'south': -82.6401,
    'aoi': 'lat_long'
}


def build_query(**kw):
    """Build a query dict for NASA's landsat search API.

    :param start gaia.core.Time: The start date
    :param end gaia.core.Time: The end date
    :param cloud float: Maximum cloud cover percentage [0,100]
    :param latitude float: A latitude covered by the image (sets south/north)
    :param longitude float: A longitude covered by the image (sets east/west)
    :param west float: The minimum longitude
    :param east float: The maximum longitude
    :param south float: The minimum latitude
    :param north float: The maximum latitude
    :param limit int: The maximum number of results to return
    :param sensor str: The landsat sensor to query

    :returns: dict
    """
    q = dict(**_defaults)
    q.update(kw)

    q['start'] = Time.format(q['start'], config.get(_section, 'datefmt'))
    q['end'] = Time.format(q['end'], config.get(_section, 'datefmt'))

    if q.get('latitude') is not None:
        q['south'] = q.pop('latitude')
        q['north'] = q['south']

    if q.get('longitude') is not None:
        q['west'] = q.pop('longitude')
        q['east'] = q['west']

    return {
        'start_date': q['start'],
        'end_date': q['end'],
        'sensor_name': q['sensor'],
        'aoi_entry': q['aoi'],
        'north_latitude': q['north'],
        'south_latitude': q['south'],
        'west_longitude': q['west'],
        'east_longitude': q['east'],
        'limit': q['limit'],
        'output_type': 'CSV'
    }


def files(**kw):
    """Search landsat files using landsat-util API.

    :param start gaia.core.Time: The start date
    :param end gaia.core.Time: The end date
    :param cloud float: Maximum cloud cover percentage [0,100]
    :param latitude float: A latitude covered by the image (sets south/north)
    :param longitude float: A longitude covered by the image (sets east/west)
    :param west float: The minimum longitude
    :param east float: The maximum longitude
    :param south float: The minimum latitude
    :param north float: The maximum latitude
    :param limit int: The maximum number of results to return
    :param sensor str: The landsat sensor to query
    """
    query = build_query(**kw)
    url = config.get(_section, 'url')
    req = requests.get(url, params=query)
    return req
